/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * "Portions Copyright (c) 2000-2001 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/FSCL.txt
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author: Burton Samograd (burton@wizard.ca)
 * CVS Id: $Id: lmsmtpd.c,v 1.69 2002/11/22 18:15:39 burton Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

/* Standard C includes */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <time.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <sys/stat.h>

/* Linux Magic library includes */
#define LM_USE_LOGGER
#include <liblm.h>

/* lmsmtpd includes */
#include "lmsmtpd.h"
#include "lmsmtpd_config.h"
#include "lmsmtp.h"
#include "cdb.h"
#include "check_user.h"
#include "smtp_checks.h"

#define LMSMTPD_LOG_SYS "lmsmtpd"

/* Main configuration data, see config.h for contents */
lmsmtpd_config_t config;

/* Module static variables */
static lm_string_t smtp_cmd;
static lm_string_t smtp_rcpt_list;
static lm_string_t smtp_from_addr;
static int smtp_message_size;
static int smtp_message_size_overflow;
static int smtp_error;
static int penv[2];
static int pdata[2];
static int childpid;
static int num_hops;
static int bad_mail_from_flag;
static char *data_return_msg;
static char *readbuf;
static int timeout_flag;
static int rcpt_delay;
static int rcpt_count;

/* Constant strings which are all allocated at the start */

static lm_string_t T = NULL;
static lm_string_t F = NULL;


static void lmsmtpd_error_out_of_memory(void);
int lmsmtpd_timeout_read(int fd, lm_string_t* line);
int lmsmtpd_write(int fd, lm_string_t line);

/* Handle resettting of all internal variables */
int lmsmtpd_rset_callback(void)
{
	int retval;

	/* Free the old data strings */
	LM_STRING_FREE(smtp_rcpt_list);
	LM_STRING_FREE(smtp_from_addr);

	/* Reset the strings to empty */
	retval = 0;
	LM_STRING_NEW(smtp_rcpt_list, "");
	retval = (smtp_rcpt_list == NULL);
	LM_STRING_NEW(smtp_from_addr, "");
	retval |= (smtp_from_addr == NULL);

	if(retval != 0)
	{
		lmsmtpd_error_out_of_memory();
	}

	/* Reset the message size and overflow flag to zero */
	smtp_message_size = 0;
	smtp_message_size_overflow = 0;

	/* Reset the error flag */
	smtp_error = 0;

	/* Reset data callback return message string */
	data_return_msg = 0;

	/* Reset the bad mail from flag */
	bad_mail_from_flag = 0;

	/* Reset the rcpt count callbacks */
	rcpt_count = 0;
	
	return 0;
}

/* Perform cleanup operations and exit */
int lmsmtpd_quit_callback(int retval)
{
	if(timeout_flag == 0)
		printf("221 %s \r\n", LM_STRING_BUFFER(config.smtpgreeting));

	/* Do cleanup operations */
	lmsmtpd_config_cleanup();
	lmsmtpd_check_user_cleanup();
	smtp_cleanup();

	if(smtp_cmd != NULL)
		LM_STRING_FREE(smtp_cmd);

	/* Free any data that has been read */
	LM_STRING_FREE(smtp_rcpt_list);
	LM_STRING_FREE(smtp_from_addr);
	LM_STRING_FREE(F);
	LM_STRING_FREE(T);

	LM_FREE(readbuf);	

	if(config.debuglevel > LM_LOG_LEVEL_WARN)
	{
		LM_MEMSTAT(LM_SMTPD_LOG_OUT);
		LM_MEMDUMP(LM_SMTPD_LOG_OUT);
	}

	fflush(stdout);
	fflush(stderr);

	exit(retval);

    /* return prevents compiler warnings even though it will never be called */
	return retval;
}

static void lmsmtpd_error_out_of_memory(void)
{
	/* Print no mem error message */
	printf("421 out of memory (#4.3.0)\r\n");
	/* Die with exit code 1 */
	lmsmtpd_quit_callback(1);
}

/* Exec the qmail-queue and setup write pipes to send the envelope and 
   body data as they are provided by the queue callback */
int lmsmtpd_queue_init(void)
{
	char *args[] = { NULL, NULL };
	char *buffer = NULL;
	int buffersize;
	int numchars;

	args[0] = LM_STRING_BUFFER(config.qmailqueue);

	/* Allocate the read buffer */
	buffersize = config.maxlinelen;
	if(LM_MALLOC(buffer, buffersize) == NULL)
	{
		LM_ASSERT(buffer == NULL);
		lmsmtpd_error_out_of_memory();
	}

	/* Reset the number of hops for new message */
	num_hops = 0;

	if( (pipe(penv) == -1) || (pipe(pdata) == -1) )
		perror("pipe error\n");

	switch(childpid = fork())
	{
		case 0:
			/* Child */

            /* close descriptor 0 and re-open it as the read end of pdata */
			close(0);
			dup(pdata[0]);

            /* close the pdata pipe (we've still got descriptor 0 pointing to it) */
			close(pdata[0]);
			close(pdata[1]);

            /* close descriptor 1 and re-open it as the read end of penv */ 
			close(1);
			dup(penv[0]);

            /* close the penv pipe (we've still got descriptor 1 pointing to it) */
			close(penv[0]);
			close(penv[1]);

            /* run qmail-queue (which reads from descriptors 0 and 1) */
			/* TODO Is this safe? Might need to reconsider executing a string that is 
			 * pulled from the environment 
			 */
			if(execvp(LM_STRING_BUFFER(config.qmailqueue), args))
			{
				perror("child: exec error");
				return -1;
			}
			_exit(1);
			break;

        case -1: 
            /* fork() error */
			LM_LOG_ERROR(LMSMTPD_LOG_SYS, "fork error");
            break;

		default:
			/* Parent fall through */
	}

	/* Write the Recieved: line from us for the message */
	numchars = snprintf(buffer, buffersize, "Received: from %s ",
			LM_STRING_BUFFER(config.tcpremotehost));
	if((smtp_error == 0) && (numchars > 0) && (write(pdata[1], buffer, numchars) != numchars))
		smtp_error |= 1;

	/* If we were given a helo domain different than tcp server, print it */
	if(config.fakehelohost != 0)
	{
		numchars = snprintf(buffer, buffersize, "(HELO %s) ",
				LM_STRING_BUFFER(config.helohost));
		if((smtp_error == 0) && (numchars > 0) && (write(pdata[1], buffer, numchars) != numchars))
			smtp_error |= 1;
	}

	/* Write an ( for the remote host information to be contained in */
	numchars = snprintf(buffer, buffersize, "(");
	if((smtp_error == 0) && (numchars > 0) && (write(pdata[1], buffer, numchars) != numchars))
		smtp_error |= 1;

	/* Write the remote info if given by tcpserver */
	if(LM_STRING_BUFFER(config.tcpremoteinfo)[0] != '\0') 
	{
		numchars = snprintf(buffer, buffersize, "%s@", LM_STRING_BUFFER(config.tcpremoteinfo));
		if((smtp_error == 0) && numchars > 0 && write(pdata[1], buffer, numchars) != numchars)
			smtp_error |= 1;
	}

	/* Write the remote ip */
	numchars = snprintf(buffer, buffersize, "%s)\n by %s with SMTP; ",
			LM_STRING_BUFFER(config.tcpremoteip),
			LM_STRING_BUFFER(config.tcplocalip));
	if((smtp_error == 0) && numchars > 0 && write(pdata[1], buffer, numchars) != numchars)
		smtp_error |= 1;

	/* Print out the time in the recieved line */
	/* TODO Not exact format as qmail
	   We do: Thu Oct 24 21:54:29 2002 -0000
	   Qmail does: 24 Oct 2002 21:54:29 -0000
	   Both are gmt tho, so that's good
	   */
	{
		time_t t;
		struct tm *times;

		time(&t);

        /* 
		 * Print out the time string according to the RFC and man strftime
         */
		times = gmtime(&t);
		numchars = strftime(buffer, buffersize, "%a, %d %b %Y %H:%M:%S %z\n", times); 
		if((smtp_error == 0) && numchars > 0 && write(pdata[1], buffer, numchars) != numchars)
			smtp_error |= 1;
	}

	/* Free the buffer */
	LM_FREE(buffer);
	return 0;
}

/* Write the envelope data to qmail-queue  */
int lmsmtpd_queue(void)
{
	int index;
	int retval;

	/* Need to close the message body pipe so that we can write to the header
	   data pipe */
	close(pdata[1]);

	/* Write the envelope to file descriptor 1 */

	/* Write the from address and it's trailing \0 to the message envelope pipe */ 
	LM_STRING_ASSERT(smtp_from_addr);
	smtp_error |= (LM_STRING_CHECK(smtp_from_addr) != 0);
	if(smtp_error == 0)
	{
		retval = write(penv[1], "F", 1);
		smtp_error |= (retval != 1);
	}

	/* Write the from addr to the envelope pipe */
	if(smtp_error == 0)
	{
		retval = write(penv[1], LM_STRING_BUFFER(smtp_from_addr),
				LM_STRING_LEN(smtp_from_addr));
		smtp_error |= (retval != LM_STRING_LEN(smtp_from_addr));
	}

	/* QMAIL */
	/* If there was an error don't write the rcpt list */
	if(smtp_error)
	{
		LM_LOG_ERROR(LMSMTPD_LOG_SYS, "error writing envelope from section!");
		return 0;
	}

	/* Find the location of the first T in the rcpt list */
	if(LM_STRING_CHR(smtp_rcpt_list, 'T', index))
	{
		/* If there isn't one, there was a serious problem */
		LM_LOG_ERROR(LMSMTPD_LOG_SYS, "empty rcpt list!!!");
		return -1;
	}

	/* index should always be 0, since all rcpt addresses start with a T */
	LM_ASSERT(index == 0);
	while(index == 0)
	{
		/* Shift off the T */
		LM_ASSERT(LM_STRING_BUFFER(smtp_rcpt_list)[0] == 'T');
		LM_STRING_START_SHIFT(smtp_rcpt_list, 1);

		/* Write a T to the pipe */
		if(smtp_error == 0)
		{
			retval = write(penv[1], "T", 1);
			smtp_error |= (retval != 1);
		}

		/* Find the next T */
		if(LM_STRING_CHR(smtp_rcpt_list, 'T', index))
		{
			/* If there wasn't another T, we are at the last address in the list,
			   so we write that and then put the two null terminators to signify the
			   end of the envelope */
			if(smtp_error == 0)
			{
				retval = write(penv[1], LM_STRING_BUFFER(smtp_rcpt_list), LM_STRING_LEN(smtp_rcpt_list));
				smtp_error |= (retval != LM_STRING_LEN(smtp_rcpt_list));
			}

			/* this is actually 3 bytes long */
			if(smtp_error == 0)
			{
				retval = write(penv[1], "\0\0", 2);
				smtp_error |= (retval != 2);
			}
			break;
		}
		else
		{
			/* We found another T, so we write the address up to where the T was found 
			   and then write a single null terminator and then shift that address off
			   the start of the string  */
			LM_ASSERT(index <= LM_STRING_LEN(smtp_rcpt_list));
			if(smtp_error == 0)
			{
				retval = write(penv[1], LM_STRING_BUFFER(smtp_rcpt_list), index);
				smtp_error |= (retval != index);
			}

			/* this is actually 2 bytes long */
			if(smtp_error == 0)
			{
				retval = write(penv[1], "\0", 1);
				smtp_error |= (retval != 1);
			}

			LM_STRING_START_SHIFT(smtp_rcpt_list, index);
			index = 0;
		}
	}

	return 0;
}

/* Close the pipes to the qmail-queue process that are still open, and wait for the 
 * child to exit and get it's exit code */ 
int lmsmtpd_queue_cleanup(void)
{
	int wpret;
	char childret;
	int retval;

	/* Close the pipe */
	close(penv[1]);

	/* Wait for the child to stop */
	do {
		wpret = waitpid(childpid, &retval, 0);
	} while (wpret == -1 && errno == EINTR);

	/* Check if there was some sort of wierd problem with waitpid */
	if(wpret != childpid)
	{
		/* TODO not tested */
		data_return_msg = "451 qq waitpid surprise (#4.3.0)\r\n";
		return 0;
	}

	/* Check to see if qmail-queue crashed */
	if(retval & 127)
	{
		/* TODO not tested */
		data_return_msg = "451 qq crashed (#4.3.0)\r\n";
	}

	/* Check the exti status of qmail queue and print the appropriate return message */
	/* I have bad feelings about WEXITSTATUS returning 0-255. 
	 * This will make sure the value is sign extended to an int
	 */
	childret = WEXITSTATUS(retval);
	retval = childret;
	if(retval > 0)
	{
		switch(retval)
		{
			case 115:
			case 11:
				data_return_msg =  "554 envelope address too long for qq (#5.1.3)";
				break;
			case 31:
				data_return_msg =  "554 mail server permanently rejected message (#5.3.0)";
				break;
			case 51:
				data_return_msg =  "451 qq out of memory (#4.3.0)";
				break;
			case 52:
				data_return_msg =  "451 qq timeout (#4.3.0)";
				break;
			case 53:
				data_return_msg =  "451 qq write error or disk full (#4.3.0)";
				break;
			case 0:
				if (!smtp_error)
				{
					break;
				}
				/* fall through if there was an error */
			case 54:
				data_return_msg =  "451 qq read error (#4.3.0)";
				break;
			case 55:
				data_return_msg =  "451 qq unable to read configuration (#4.3.0)";
				break;
			case 56:
				data_return_msg =  "451 qq trouble making network connection (#4.3.0)";
				break;
			case 61:
				data_return_msg =  "451 qq trouble in home directory (#4.3.0)";
				break;
			case 63:
			case 64:
			case 65:
			case 66:
			case 62:
				data_return_msg =  "451 qq trouble creating files in queue (#4.3.0)";
				break;
			case 71:
				data_return_msg =  "451 mail server temporarily rejected message (#4.3.0)";
				break;
			case 72:
				data_return_msg =  "451 connection to mail server timed out (#4.4.1)";
				break;
			case 73:
				data_return_msg =  "451 connection to mail server rejected (#4.4.1)";
				break;
			case 74:
				data_return_msg =  "451 communication with mail server failed (#4.4.2)";
				break;
			case 91: /* fall through */
			case 81:
				data_return_msg =  "451 qq internal bug (#4.3.0)";
				break;
			case 120:
				data_return_msg =  "451 unable to exec qq (#4.3.0)";
				break;
			default:
				if ((retval >= 11) && (retval <= 40))
				{
					data_return_msg =  "554 qq permanent problem (#5.3.0)";
					break;
				}
				data_return_msg =  "451 qq temporary problem (#4.3.0)";
				break;
		}
	}

	return retval > 0;
}

/* Check if this message has been bounced or relayed and increment the hop
   count if it has */
void lmsmtpd_header_check_hop(lm_string_t line)
{
	if((toupper(LM_STRING_BUFFER(line)[0]) == 'R' &&
				toupper(LM_STRING_BUFFER(line)[1]) == 'E' &&
				toupper(LM_STRING_BUFFER(line)[2]) == 'C' &&
				toupper(LM_STRING_BUFFER(line)[3]) == 'I' &&
				toupper(LM_STRING_BUFFER(line)[4]) == 'E' &&
				toupper(LM_STRING_BUFFER(line)[5]) == 'V' &&
				toupper(LM_STRING_BUFFER(line)[6]) == 'E' &&
				toupper(LM_STRING_BUFFER(line)[7]) == 'D') ||
			(toupper(LM_STRING_BUFFER(line)[0]) == 'D' &&
			 toupper(LM_STRING_BUFFER(line)[1]) == 'E' &&
			 toupper(LM_STRING_BUFFER(line)[2]) == 'L' &&
			 toupper(LM_STRING_BUFFER(line)[3]) == 'I' &&
			 toupper(LM_STRING_BUFFER(line)[4]) == 'V' &&
			 toupper(LM_STRING_BUFFER(line)[5]) == 'E' &&
			 toupper(LM_STRING_BUFFER(line)[6]) == 'R' &&
			 toupper(LM_STRING_BUFFER(line)[7]) == 'E' &&
			 toupper(LM_STRING_BUFFER(line)[8]) == 'D'))
	{
		LM_LOG_MSG(LMSMTPD_LOG_SYS, "incrementing hop count");
		num_hops++;

	}
}

int lmsmtpd_update_msg_size(int size)
{
	unsigned int tmpsize;

	/* Add the length of the line to the total message size */
	/* Check for overflow of smtp_message_size */
	tmpsize = smtp_message_size;
	tmpsize += size;

	/* If the value of tmpsize is less than the old message size,
	 * there was a size overflow so we'll flag an error
	 */
	if((unsigned int)tmpsize > (unsigned int)smtp_message_size)
	{
		smtp_message_size = tmpsize;
		if(config.databytes != 0)
			if(smtp_message_size  > config.databytes)
			{
				smtp_error |= 1;
				return -1;
			}
	}
	else
	{
		/* Overflow in size variable error */
		smtp_message_size_overflow = 1;
		smtp_error |= 1;
	}

	return 0;

}

/* This function is called for each line contained in the header
   Simply writes it to the qmail-queue message pipe for now */
int lmsmtpd_header_line_callback(lm_string_t line)
{
	int retval;

	/* Update the total number of bytes that have been recieved */
	lmsmtpd_update_msg_size(LM_STRING_LEN(line));

	/* Perform hop checking like qmail does */
	lmsmtpd_header_check_hop(line);

	/* If there hasn't been an error yet, write the line to qmail-queue */
	if(smtp_error == 0)
		retval = write(pdata[1], LM_STRING_BUFFER(line), LM_STRING_LEN(line));

	/* Flag an error if line wasn't written fully */
	smtp_error |= (retval != LM_STRING_LEN(line));

	return 0;
}

/* This function is called for each line contained in the header
   Simply writes it to the qmail-queue message pipe for now */
int lmsmtpd_body_line_callback(lm_string_t line)
{
	int retval;

	/* Update the total number of bytes that have been recieved */
	lmsmtpd_update_msg_size(LM_STRING_LEN(line));

	/* Write the line to the data pipe */
	if(!smtp_error)
		retval = write(pdata[1], LM_STRING_BUFFER(line), LM_STRING_LEN(line));

	/* Flag an error if line wasn't written fully */
	smtp_error |= (retval != LM_STRING_LEN(line));

	return 0;
}

/* Function called after all queuing functions have completed */
int lmsmtpd_data_callback(void)
{
	/* If the number of hops was too great print warning */
	if(num_hops > config.maxhops)
	{
		LM_LOG_MSG(LMSMTPD_LOG_SYS, "max hop count exceeded!");

		/* Flag an error for to be caught later */
		smtp_error |= 1;

		printf("554 too many hops, this message is looping (#5.4.6)\r\n");
		return 0;
	}

	/* Check if the message size is greater than maximum size and print 
	 * error return code if
	 * Also check if there was an overflow.
	 */
	if((smtp_message_size > config.databytes) || smtp_message_size_overflow)
	{
		LM_LOG_MSG(LMSMTPD_LOG_SYS, "message size greater than databytes!");
		printf("552 sorry, that message size exceeds my databytes limit (#5.3.4)\r\n");
		return 0;
	}

	/* Print the current time and the pid of the child process for return
	   code message string */
	if(data_return_msg)
		printf("%s\r\n", data_return_msg);
	else
		printf("250 ok %d qp %d\r\n", (int)time(NULL), childpid);

	return 0;
}

/* Check to see if the given address is in the bad mail from list */
int lmsmtpd_bad_mail_from(lm_string_t addr)
{
	lm_string_t bmf = NULL;
	int index = -1;
	int bmf_flag = 0;

	/* Check for bad mail status from given address and return
	   zero if the address is good, or non-zero if the address is not 
	   allowed by badmailfrom */
	if(LM_STRING_LEN(config.badmailfrom) == 0)
		return 0;

	/* Make a copy of the list so we can shift it */
	if(LM_STRING_DUP(bmf, config.badmailfrom))
	{
		lmsmtpd_error_out_of_memory();
	}

	/* This list of address is seperated by \n's. We'll go through 
	 * the list  and create a new string for each entry in the list 
	 * and then shift the bmf string over to get that one
	 off the list.  This will exit when the bmf list is empty */
	while((LM_STRING_BUFFER(bmf)[0] != '\0') && (bad_mail_from_flag == 0))
	{
		lm_string_t line = NULL;

		if(LM_STRING_GET_LINE(bmf, line))
		{
			lmsmtpd_error_out_of_memory();
		}

		/* Now we check to see if this is an @host address or user@host */
		if(LM_STRING_BUFFER(line)[0] == '@')
		{
			/* We have a group host address */
			/* Shift off the username on the given address so we
			 * can do a comparison */
			if(LM_STRING_CHR(addr, '@', index) == 0)
			{
				lm_string_t addrcpy = NULL;

				if(LM_STRING_DUP(addrcpy, addr) != 0)
				{
					lmsmtpd_error_out_of_memory();
				}
				LM_STRING_START_SHIFT(addrcpy, index);
				if(LM_STRING_CMP(addrcpy, line) == 0)
				{
					bmf_flag = 1;
				}
				LM_STRING_FREE(addrcpy);
			}
			/* We were just passed a usename with no @, so we can't do the check. */
		}
		else
		{
			/* We have a single address */
			if(LM_STRING_CMP(line, addr) == 0)
			{
				/* We found a match, set bmf flag */
				bmf_flag = 1;
			}
		}

		/* Free this address and go onto the next */
		LM_STRING_FREE(line);
	}
	LM_STRING_FREE(bmf);

	return bmf_flag;
}

/* Function called after all queuing functions have completed */
int lmsmtpd_mail_callback(lm_string_t addr)
{
	/* Perform a reset when this is called */
	lmsmtpd_rset_callback();

	/* Check the address to see if it's in the badmailfrom list */
	if(lmsmtpd_bad_mail_from(addr) != 0)
	{
		bad_mail_from_flag = 1;
		return 0;
	}

	/* Print an ok message if everything was ok */
	printf("250 ok\r\n");

	/* Save the parsed address */
	LM_STRING_FREE(smtp_from_addr);
	if(LM_STRING_DUP(smtp_from_addr, addr) != 0)
	{
		lmsmtpd_error_out_of_memory();
	}

	return 1;
}

/* Check the given hostname to see if it is contained in the rcpthosts 
 * file and we should accept the mail 
 */
int lmsmtpd_rcpt_host(lm_string_t addr)
{
	lm_string_t rcpthosts = NULL;
	lm_string_t rcptaddr = NULL;
	int index;

	/* Make copies of the address and rcpthosts string so we can modify them */
	if(LM_STRING_DUP(rcpthosts, config.rcpthosts) != 0) 
	{
		lmsmtpd_error_out_of_memory();
	}

	/* See if there is an @ sign in the address parameter.
	   If not, then we have a local address, so the domain we're
	   going to check for is localhost */
	if(LM_STRING_CHR(addr, '@', index) != 0)
	{
		/* No @ sign so we have a local address */
		if(LM_STRING_NEW(rcptaddr, "localhost") != 0)
		{
			lmsmtpd_error_out_of_memory();
		}
	}
	else
	{
		/* Shift off the user name so we only have the domain */
		if(LM_STRING_DUP(rcptaddr, addr) != 0)
		{
			lmsmtpd_error_out_of_memory();
		}
		LM_STRING_START_SHIFT(rcptaddr, index+1);
	}

	/* Reject all addresses which start with a leading . */
	if(LM_STRING_BUFFER(rcptaddr)[0] == '.')
	{
		LM_STRING_FREE(rcpthosts);
		LM_STRING_FREE(rcptaddr);
		return -1;
	}

	/* For each of the lines contained in the rcpthosts variable, 
	   convert the line to a string and then compare that to the 
	   domain from the given address.  If the address matches one of
	   the rcpt domains, then return success */
	while(LM_STRING_LEN(rcpthosts) != 0)
	{
		lm_string_t domain = NULL;

		if(LM_STRING_GET_LINE(rcpthosts, domain) != 0)
		{
			lmsmtpd_error_out_of_memory();
		}

		/* See if the line from the rcpthosts and the given 
		 * domain match 
		 */
		if(!smtp_match_domain(rcptaddr, domain))
		{
			/* Free the allocated strings */
			LM_STRING_FREE(rcptaddr);
			LM_STRING_FREE(rcpthosts);
			LM_STRING_FREE(domain);

			/* They did, so return success */
			return 0;
		}
		LM_STRING_FREE(domain);
	}
	LM_STRING_FREE(rcptaddr);
	LM_STRING_FREE(rcpthosts);
	return -1;
}


int lmsmtpd_do_smtp_checks(spam_rule_config_t *rules, lm_string_t addr)
{
	lm_string_t fromuser = NULL;
	lm_string_t fromdomain = NULL;
	int index;
	int retval = 0;

	LM_ASSERT(rules != NULL);
	LM_STRING_ASSERT(addr);
	LM_ASSERT(rules->smtp_checks != 0);

	/* Split the from address into user and domain parts */
	LM_STRING_DUP(fromuser, smtp_from_addr);
	if(!LM_STRING_CHR(fromuser, '@', index))
	{
		LM_STRING_DUP(fromdomain, smtp_from_addr);
		LM_STRING_END_SHIFT(fromuser, LM_STRING_LEN(fromuser)-index);
		LM_STRING_START_SHIFT(fromdomain, index+1);
	}
	else
	{
		/* There was no @ specified so set the domain to nothing */
		LM_STRING_NEW(fromdomain, "");
	}

	LM_LOG_MSG(LMSMTPD_LOG_SYS, "Performing SMTP based checks...");


	/* Check to see if the sender provided a resolvable helo domain */
	if((rules->valid_helo_domain != 0) &&
			(retval == 0) &&
			(LM_STRING_LEN(config.helohost) > 0))
		if((retval |= smtp_check_valid_helo_domain(config.helohost) != 0))
		{
			LM_LOG_WARN(LMSMTPD_LOG_SYS, "HELO command domain does not resolve!");
		}

	/* Check to see if the sender provided a valid domain in thier from address */
	if((rules->from_valid_domain != 0) &&
			(retval == 0) &&
			(LM_STRING_LEN(fromdomain) > 0))
		if((retval |= smtp_check_from_valid_domain(fromdomain) != 0))
		{
			LM_LOG_WARN(LMSMTPD_LOG_SYS, "From: domain does not resolve!");
		}

	/* Check to see if the sender provided a valid domain in thier from address */
	if((rules->accept_mail_from_self != 0) &&
			(retval == 0))
		if((retval |= smtp_check_mail_from_self(addr, smtp_from_addr) != 0))
		{
			LM_LOG_WARN(LMSMTPD_LOG_SYS, "Address in MAIL command identical to RCPT command!");
		}

	LM_STRING_FREE(fromuser);
	LM_STRING_FREE(fromdomain);
	return retval;
}

int lmsmtpd_rcpt_callback(lm_string_t addr)
{
	int retval;
	int spam_check_retval;
	int index;
	lm_string_t dotspam = NULL;

	/* Increment the number of rcpts that we've recieved */
	rcpt_count++;

	/* Check how many rcpt's we've gotten and do a delay if over given value */
	if(rcpt_count > config.rcptdelayat)
	{
		if(rcpt_delay < config.rcptdelaymax)
			rcpt_delay += config.rcptdelayinc;
		sleep(rcpt_delay);
	}

	/* Check how many rcpt to's we've gotten and return error if 
	 * 	over the max value given */
	if(rcpt_count > config.maxrcpt)
	{
		/* print temp error message and continue */
		printf("452 Too many recipients (#4.5.3.1)\r\n");

		/* Prevent an overflow from a very determined sender */
		rcpt_count = config.maxrcpt;

		return 0;
	}

	/* If the sender was in the bad mail from list, print an error mesage and return */
	if(bad_mail_from_flag)
	{
		printf("553 sorry, your envelope sender is in my badmailfrom list (#5.7.1)\r\n");
		return -1;
	}


	/* If the relayclient variable was set by tcp server */
	if(LM_STRING_LEN(config.tcprelayclient) > 0)
	{
		/* Append the value of relayclient after the address */
		if(LM_STRING_CAT(addr, config.tcprelayclient) != 0)
		{
			lmsmtpd_error_out_of_memory();
		}
	}
	else
	{
		/* Check to see if we have a local address
		 * (no @ sign)
		 */
		if(LM_STRING_CHR(addr, '@', index) == 0)
		{
			/* TODO Check for local ip address after @
			 * sign that matches machine ip, which would
			 * also be local delivery
			 * */

			/*  Check to see if the domain is contained in rcpthosts */
			if(LM_STRING_LEN(config.rcpthosts) > 0)
			{
				if(lmsmtpd_rcpt_host(addr) != 0)
				{
					/* Return error so that the caller knows that this wasn't
					   a good rcpt */
					printf("553 sorry, that domain isn't in my list of allowed rcpthosts (#5.7.1)\r\n");
					return -1;
				}
			}
		}

		/*  Check to see if the rcpt address is a valid user and
			find their information */
		/* Per user checking has been turned into a configuration variable, so 
		 * only check if it is set.  */
		retval = 0;
		if(config.check_valid_users != 0)
		{
			retval = lmsmtpd_check_rcpt_user(&addr, &dotspam);
		}

		if(retval > 0)
		{
			char buffer[1024];

			/* Output user does not exist */
			printf("550 User does not exist\r\n");

			/* Log message about denying mail */
			snprintf(buffer, sizeof(buffer),
					"%s BLOCKED as non-valid user", LM_STRING_BUFFER(addr));
			LM_LOG_WARN(LMSMTPD_LOG_SYS, buffer);

			/* Return error so that rcpt counter is not
			   incremented */
			if(dotspam)
				LM_STRING_FREE(dotspam);

			return -1;
		}
		else if( retval < 0)
		{
			/* There was some error with checking the user, 
			   but still pass the mail through to qmail 
			   */
			LM_LOG_ERROR("lmsmptd", "error checking for user validity");
		}
		/* User was valid */

		/* We know that the user was valid.  If dotspam has a non-null value,
		   the user has a file which matches the address and want's spam checking
		   on that address.  We should load the spam configuration variables now */
		if(dotspam != NULL)
		{
			spam_rule_config_t spam_rules;

			memset(&spam_rules, 0, sizeof(spam_rules));
			/* Load the spam rules */
			retval = lmsmtpd_config_load_spam_rules(&spam_rules, dotspam);
			if(retval == 0)
			{
				/* Spam rules loaded ok */
				LM_LOG_MSG(LMSMTPD_LOG_SYS, "User spam rules loaded sucessfully");


				/* Check to see if the user has spam checking enabled */
				if(spam_rules.spam_check != 0)
				{
					LM_LOG_MSG(LMSMTPD_LOG_SYS, "User spam checking enabled.");

					/* Perform checks based on smtp information */
					if(spam_rules.smtp_checks != 0)
					{
						spam_check_retval = lmsmtpd_do_smtp_checks(&spam_rules, addr);
					}
				}
				else
				{
					LM_LOG_MSG(LMSMTPD_LOG_SYS, "User does not have spam checking enabled.");
				}

				if(lmsmtpd_config_free_spam_rules(&spam_rules))
					return -1;
			}
			else
			{
				/* There was an error loading the spam rules */
				LM_LOG_MSG(LMSMTPD_LOG_SYS, "Failed to load user spam rules");
			}
		}
	}

	if(dotspam != NULL)
		LM_STRING_FREE(dotspam);

	/* If the return value from the initial spam checking does not suceeed, don't 
	 * add the recipient to the envelope from list
	 */
	if(spam_check_retval == 0)
	{
		/* Concat "T" + address to rcpt list string */
		if((LM_STRING_CAT(smtp_rcpt_list, T) != 0) ||
				(LM_STRING_CAT(smtp_rcpt_list, addr) != 0))
		{
			lmsmtpd_error_out_of_memory(); 
		}
	}
	else
	{
		/* TODO This code needs to be tested */

		/* Send the mail to user-$LM_SMTPD_QUARANTINE_EXT@domain if quarantine enabled */
		if(config.quarantine_on_hit != 0)
		{
			lm_string_t tmpdomain = NULL;
			lm_string_t tmpaddr = NULL;
			int index;

			/* TODO Needs to add quarantine string before the @ sign of the address */
			LM_STRING_DUP(tmpdomain, addr);
			LM_STRING_DUP(tmpaddr, addr);

			if(!LM_STRING_CHR(tmpaddr, '@', index))
			{
				LM_STRING_END_SHIFT(tmpaddr, LM_STRING_LEN(tmpaddr)-index);
				LM_STRING_START_SHIFT(tmpdomain, index);
				LM_STRING_CAT(tmpaddr, config.quarantine_ext);
				LM_STRING_CAT(tmpaddr, tmpdomain);
			}
			else
			{
				/* Big error */
				return -1;
			}

			LM_ASSERT(config.quarantine_ext != NULL);
			if(LM_STRING_CAT(smtp_rcpt_list, T) ||
					LM_STRING_CAT(tmpaddr, config.quarantine_ext) ||
					LM_STRING_CAT(smtp_rcpt_list, tmpaddr))
			{
				lmsmtpd_error_out_of_memory();
			}
			LM_LOG_WARN(LMSMTPD_LOG_SYS, "MAIL QUARANTINED!!! ADDRESS GIVEN LM_SMTPD_QUARANTINE_EXT AND DELIVERED!");

			LM_STRING_FREE(tmpdomain);
			LM_STRING_FREE(tmpaddr);
		}

		/* Return a user does not exist message if requested */
		if(config.ret_udne_on_hit)
		{
			/* Output user does not exist */
			printf("550 User does not exist\r\n");

			/* Don't print default 250 return message */
			return -1;
		}
	}

	return 0;
}

int lmsmtpd_helo_callback(lm_string_t domain)
{
	/* Free the old helo host string and copy the arguents into a new one */
	LM_STRING_FREE(config.helohost);
	if(LM_STRING_DUP(config.helohost, domain) != 0)
	{
		lmsmtpd_error_out_of_memory();
	}

	/* Set the fakehelohost variable to show that the given helo host doesn't match
	   the one provided by tcp server */ 
	/* Returns 0 on match, fakehelohost set to non zero if they don't */
	config.fakehelohost = LM_STRING_CASECMP(config.tcpremotehost, config.helohost);

	printf("250 %s \r\n", LM_STRING_BUFFER(config.smtpgreeting));

	return 0;
}

int lmsmtpd_ehlo_callback(lm_string_t domain)
{
	/* Free the old helo host string and copy the arguents into a new one */
	LM_STRING_FREE(config.helohost);
	if(LM_STRING_DUP(config.helohost, domain) != 0)
	{
		lmsmtpd_error_out_of_memory();
	}

	/* Set the fakehelohost variable to show that the given helo host doesn't match
	   the one provided by tcp server */ 
	/* Returns 0 on match, fakehelohost set to non zero if they don't */
	config.fakehelohost = LM_STRING_CASECMP(config.tcpremotehost, config.helohost);

	printf("250 %s \r\n", LM_STRING_BUFFER(config.smtpgreeting));

	return 0;
}

/* Print a specific help string for this program */
int lmsmtpd_help_callback(void)
{
	fprintf(stdout, "214  qmail home page: http://pobox.com/~djb/qmail.html, LinuxMagic Support http://www.linuxmagic.com\r\n");
	fflush(stdout);
	return 1;
}

int lmsmtpd_init(int argc, char** argv)
{
	int retval;

	/* Perform setup by reading config */
	if(lmsmtpd_config_load(argc, argv) != 0)
	{
		LM_LOG_ERROR(LM_SMTPD_LOG_SYS, "config read error...EXITING");
		return -1;
	}

	/* Put a log message in there */
	LM_LOG_MSG(LM_SMTPD_LOG_SYS, "LinuxMagic SMTPD started");

	/* TODO proper error checking and handling for out of memory */

	/* Allocate empty strings for data that can be recieved */
	retval = 0;
	LM_STRING_NEW(smtp_rcpt_list, "");
	retval = (smtp_rcpt_list == NULL);
	LM_STRING_NEW(smtp_from_addr, "");
	retval |= (smtp_from_addr == NULL);
	LM_STRING_NEW(F, "F");
	retval |= (F == NULL);
	LM_STRING_NEW(T, "T");
	retval |= (T == NULL);

	/* Check for any memory allocation errors */
	if(retval != 0)
		lmsmtpd_error_out_of_memory();

	/* TODO Build the bad mail from list from list read from configuration */
	/* TODO Build the relay hosts list form the list read from configuration */
	/* These are both optimizations that probably won't be done until later */

	/* Allocate the read buffer */
	if(LM_MALLOC(readbuf, config.maxlinelen) == NULL)
	{
		lmsmtpd_error_out_of_memory();
	}

	/* Initialize the user checking code */
	if(lmsmtpd_check_user_init() != 0)
		lmsmtpd_error_out_of_memory();

	/* If anything failed in this we would have exited */
	return 0;
}

int main(int argc, char **argv)
{
	int retval;

	/* Initalize */
	if((retval = lmsmtpd_init(argc,argv)) != 0)
		return lmsmtpd_quit_callback(-1);

	/* Print the smtp greeting */
	printf("220 %s ESMTP\r\n", LM_STRING_BUFFER(config.smtpgreeting));
	fflush(stdout);

	/* Disable pipe signals */
	signal(SIGPIPE, SIG_IGN);

	/* Read an smtp command from stdin */
	/* This function allocated the string to hold the command */
	if(smtp_init(config.timeoutsmtpd, 
				lmsmtpd_timeout_read, lmsmtpd_write, /* read, write */
				lmsmtpd_queue_init, lmsmtpd_queue, lmsmtpd_queue_cleanup,
				lmsmtpd_helo_callback, lmsmtpd_ehlo_callback, /* helo, ehlo */
				lmsmtpd_mail_callback, lmsmtpd_rcpt_callback, /* mail, rcpt  */
				lmsmtpd_data_callback, NULL, /* data, auth */
				NULL, lmsmtpd_rset_callback, /* vrfy, rset */
				lmsmtpd_help_callback, NULL, /* help, noop */
				lmsmtpd_quit_callback, /* quit */
				lmsmtpd_header_line_callback,
				lmsmtpd_body_line_callback) != 0) /* header, body line callbacks */
	{
		return lmsmtpd_quit_callback(-1);
	}

	/* Command loop */
	while(smtp_read_command(&smtp_cmd) >= 0)
	{
		/* Parse and handle the command, checking for errors */
		/* This function frees the string holding the commmand */
		if(smtp_handle_command(smtp_cmd) != 0)
		{
			/* QMAIL */
			/* TODO Print proper message here instead of what qmail does? */
			puts("qmail-inject: fatal: read error");
			exit(111);
		}
		LM_STRING_FREE(smtp_cmd);
	}

	/* Exit the program */
	return lmsmtpd_quit_callback(0);
}

/* SIGALRM handler, prints timeout message and exits the program */
void lmstpd_sigalrm(int arg)
{
	timeout_flag = 1;
	printf("451 timeout (#4.4.2\r\n");
	lmsmtpd_quit_callback(1);
}

/* Perform a read from the given file descriptor with a timeout */
int lmsmtpd_timeout_read(int fd, lm_string_t* line)
{
	lm_string_t newstr = NULL;
	int offset;
	int retval;
	char c;

	offset = 0;
	while(1)
	{
		/* Setup an alarm signal handler */
		timeout_flag = 0;
		signal(SIGALRM, lmstpd_sigalrm);

		/* Enable the for the timeout */
		alarm(config.timeoutsmtpd);

		/* Attempt to read from the file, ignoring errors except for the alarm signal */
		while(((retval = read(fd, &c, 1)) == -1) && (errno != EINTR))
			; /* do nothing */

		if(offset < config.maxlinelen)
		{
			readbuf[offset] = c;
			LM_MEMASSERT(readbuf);
			if(c == '\n')
			{
				readbuf[offset+1] = '\0';
				break;
			}
			offset++;
		}
		else
		{
			/* Line to long, , read the rest of the line and discard until we find */
			/* a newline and then return error */
			while(c != '\n')
			{
				while(((retval = read(fd, &c, 1)) == -1) && (errno != EINTR));
				LM_MEMASSERT(readbuf);
			}
			return -1;
		}
	}

	LM_MEMASSERT(readbuf);
	if((retval > 0) && (retval < config.maxlinelen))
	{
		if(LM_STRING_NEW(newstr, readbuf) != 0)
		{
			lmsmtpd_error_out_of_memory();
		}
	}

	*line = newstr;
	return *line == NULL;
}

/* Write the given string to the file descriptor */
/* TODO Add timeouts and other error checking ? */
int lmsmtpd_write(int fd, lm_string_t line)
{
	int retval;

	/* TODO should probably do some error checking here */
	LM_STRING_ASSERT(line);
	if(LM_STRING_CHECK(line) != 0)
		return -1;

	LM_ASSERT(fd > 0);
	retval = write(fd, LM_STRING_BUFFER(line), LM_STRING_LEN(line));

	/* TODO Do we want an fsync here all the time? */
	fsync(fd);
	return retval == LM_STRING_LEN(line);
}
